#include <sys/types.h>

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>

#include "fdevent.h"
#include "settings.h"
#include "buffer.h"

/**
 * 
 *  this eventhandler isn't working nor done
 * 
 */

#if 0 && defined USE_FREEBSD_KQUEUE

#define FDEVENT_FREEBSD_KQUEUE_MAX_FDS 128

/**
 * finished
 */
static void fdevent_freebsd_kqueue_free(fdevents *ev) {
	close(ev->kqueue_fd);
}

/**
 * everything that is not marked as 'finished' is still unusable
 */

static int fdevent_freebsd_kqueue_event_del(fdevents *ev, int fde_ndx, int fd) {
	struct kevent ke;
	
	if (fde_ndx == -1) return -1;
	
	memset(&ke, 0, sizeof(ke));
	
	ke.ident = fd;
	ke.filter = EVFILT_READ;
	ke.flags = EV_DELETE;
	
	if (0 != epoll_ctl(ev->epoll_fd, EPOLL_CTL_DEL, fd, &ep)) {
		fprintf(stderr, "%s.%d: epoll_ctl failed: %s, dying\n", __FILE__, __LINE__, strerror(errno));
		
		SEGFAULT();
		
		return 0;
	}
	
	
	return -1;
}

static int fdevent_freebsd_kqueue_event_add(fdevents *ev, int fde_ndx, int fd, int events) {
	struct epoll_event ep;
	int add = 0;
	
	if (fde_ndx == -1) add = 1;
	
	ep.events = 0;
	
	if (events & FDEVENT_IN)  ep.events |= EPOLLIN;
	if (events & FDEVENT_OUT) ep.events |= EPOLLOUT;
	
	ep.events |= EPOLLERR | EPOLLHUP | EPOLLET;
	
	ep.data.ptr = NULL;
	ep.data.fd = fd;
	
	if (0 != epoll_ctl(ev->epoll_fd, add ? EPOLL_CTL_ADD : EPOLL_CTL_MOD, fd, &ep)) {
		fprintf(stderr, "%s.%d: epoll_ctl failed: %s, dying\n", __FILE__, __LINE__, strerror(errno));
		
		SEGFAULT();
		
		return 0;
	}
	
	return fd;
}

static int fdevent_freebsd_kqueue_poll(fdevents *ev, int timeout_ms) {
	int ret;
	struct timespec ts;
	
	ts.tv_sec = timeout_ms / 1000;
	ts.tv_nsec = (timeout_ms % 1000) * 1000;
	
	ret = kevent(ev->kqueue_fd, FDEVENT_FREEBSD_KQUEUE_MAX_FDS, 
		     &(ev->kqueuefds), 0, NULL, &ts);
	
	return ret;
}

static int fdevent_freebsd_kqueue_event_get_revent(fdevents *ev, size_t ndx) {
	int events = 0, e;
	
	e = ev->epoll_events[ndx].events;
	if (e & EPOLLIN) events |= FDEVENT_IN;
	if (e & EPOLLOUT) events |= FDEVENT_OUT;
	if (e & EPOLLERR) events |= FDEVENT_ERR;
	if (e & EPOLLHUP) events |= FDEVENT_HUP;
	if (e & EPOLLPRI) events |= FDEVENT_PRI;
	
	return e;
}

static int fdevent_freebsd_kqueue_event_get_fd(fdevents *ev, size_t ndx) {
	return ev->epoll_events[ndx].data.fd;
}

static int fdevent_freebsd_kqueue_event_next_fdndx(fdevents *ev, int ndx) {
	size_t i;
	
	i = (ndx < 0) ? 0 : ndx + 1;
	
	return i;
}

/**
 * finished
 */

int fdevent_freebsd_kqueue_init(fdevents *ev, size_t maxfds) {
	ev->type = FDEVENT_HANDLER_LINUX_SYSEPOLL;
#define SET(x) \
	ev->x = fdevent_freebsd_kqueue_##x;
	
	SET(free);
	SET(poll);
	
	SET(event_del);
	SET(event_add);
	
	SET(event_next_fdndx);
	SET(event_get_fd);
	SET(event_get_revent);
	
	if (-1 == (ev->kqueue_fd = kqueue())) {
		fprintf(stderr, "%s.%d: kqueue failed (%s), try to set server.event-handler = \"poll\" or \"select\"\n",
			__FILE__, __LINE__, strerror(errno));
		return -1;
	}
	
	return 0;
}

#else
int fdevent_freebsd_kqueue_init(fdevents *ev, size_t maxfds) {
	UNUSED(ev);
	UNUSED(maxfds);

	fprintf(stderr, "%s.%d: freebsd-kqueue not supported, try to set server.event-handler = \"poll\" or \"select\"\n",
		__FILE__, __LINE__);
	
	return -1;
}
#endif
